/*  Implementation of class galimber which implements rotation and projection of
 *  an object onto a plane such that hidden lines are eliminated.
 */
#include "galimber.h"
#include <HardwareSerial.h> //Support debugging prints

#define dabs(value) (((value)>0)?(value):-(value))
#define dmax(v1,v2) (((v1)>(v2))?(v1):(v2))
#define dmin(v1,v2) (((v1)<(v2))?(v1):(v2))

//Constructor
galimber::galimber(int numpin, double pin[][3], int numfin, double facein[][20]) //Constructor
{
  debug = false; //Setting this true generates tons of printout
  maxdim = 0; //This will grow as object is rotated to various orientations

  //Copy in the object description.
  nump = numpin; //Points
  for(int i=0; i<nump; i++) for(int j=0; j<3; j++)
  {
    p[i][j] = pin[i][j];
  }
  numf = numfin; //Faces
  for(int i=0; i<numf; i++) for(int j=0; j<20; j++)
  {
    face[i][j] = facein[i][j];
  }

  //Compute edge indices.
  int i,j,k,m; //Persistent loop indices
  nume=0;
  for(i=0;i<numf;i++)
  {
    for(j=0;j<18;j++)
    {
      if(face[i][j+1]!=EOF)
      {
        if(face[i][j+1]>=face[i][j])
        {
          if(nume<200)
          {
            edgeno[nume][0]=face[i][j];
            edgeno[nume][1]=face[i][j+1];
            nume++;
          }
          else
          {
            error("1005");
          }
        }
      }
      else
      {
        break;
      }
    }
    if(face[i][0]>=face[i][j])
    {
      if(nume<200)
      {
        edgeno[nume][0]=face[i][j];
        edgeno[nume][1]=face[i][0];
        nume++;
      }
      else
      {
        error("1005");
      }
    }
  }

  //Compute edge-face relationship.
  for(i=0;i<nume;i++)
  {
    icopla=0;
    edgefa[i][0]=edgefa[i][1]=EOF;
    for(j=0;j<numf;j++)
    {
      if(face[j][19])icopla++;
      for(k=0;k<18;k++)
      {
        if(face[j][k+1]!=EOF)
        {
          if(face[j][k]==edgeno[i][0] &&
             face[j][k+1]==edgeno[i][1])
          {
            edgefa[i][0]=(short)(j-icopla);
          }
          if(face[j][k+1]==edgeno[i][0] &&
             face[j][k]==edgeno[i][1])
          {
            edgefa[i][1]=(short)(j-icopla);
          }
        }
        else
        {
          break;
        }
      }
      if(face[j][k]==edgeno[i][0] && face[j][0]==edgeno[i][1])
      {
        edgefa[i][0]=(short)(j-icopla);
      }
      if(face[j][0]==edgeno[i][0] && face[j][k]==edgeno[i][1])
      {
        edgefa[i][1]=(short)(j-icopla);
      }
    }
  }

  //Compute inward unit normals to faces.
  {
    register int index,jndex,kndex;
    register int jj;
    double vec1[3],vec2[3];
    register double amag;
    icopla=0;
    for(k=0;k<numf;k++)
    {
      if(face[k][19])
      {
        icopla++;
      }
      else
      {
        jj=k-icopla;
        index=face[k][0];
        jndex=face[k][1];
        kndex=face[k][2];
        for(j=0;j<3;j++)
        {
          vec1[j]=p[jndex][j]-p[index][j];
          vec2[j]=p[kndex][j]-p[jndex][j];
        }
        for(i=0;i<3;i++)
        {
          normal[jj][i]=
              vec1[(i+1)%3]*vec2[(i+2)%3]-
              vec1[(i+2)%3]*vec2[(i+1)%3];
        }
        amag=sqrt(normal[jj][0]*normal[jj][0]+
                  normal[jj][1]*normal[jj][1]+
                  normal[jj][2]*normal[jj][2]);
        for(i=0;i<3;i++)
        {
          normal[jj][i]/=amag;
        }
      }
    }
  }

  //Compute concavity factors for edges.
  for(i=0;i<nume;i++)
  {
    concav[i]=0.;
    for(j=0;j<3;j++)
    {
      concav[i]+=(p[edgeno[i][1]][j]-p[edgeno[i][0]][j])*
          (normal[edgefa[i][0]][(j+1)%3]*
           normal[edgefa[i][1]][(j+2)%3]-
           normal[edgefa[i][0]][(j+2)%3]*
           normal[edgefa[i][1]][(j+1)%3]);
    }
  }

  //Print out processed object description
  Serial.printf("The %d object nodes:\n\nNode #\t   x\t   y\t   z\n",nump);
  for(i=0;i<nump;i++)
  {
    Serial.printf("%4d\t%7.2f\t%7.2f\t%7.2f\n",i,p[i][0],p[i][1],p[i][2]);
  }

  Serial.printf("\n\nThe %d object faces:\n\nFace #\t  Set\t   Right-handed-node-list\n",
    numf-icopla);
  icopla=0;
  for(i=k=0;i<numf;i++)
  {
    for(j=0;j<19;j++)
    {
      if(face[i][j]==EOF)break;
    }
    if(face[i][19])
    {
      icopla++;
      k++;
    }
    else
    {
      k=0;
    }
    Serial.printf("%4d\t%4d\t",i-icopla,k);
    for(m=0;m<j;m++)
    {
      Serial.printf("%4d",face[i][m]);
    }
    Serial.printf("\n");
  }

  Serial.printf("\n\nGalimbr has computed %d object edges:\n\n",nume);
  Serial.printf("Edge #\tFrom\tTo\tRight\tLeft\tEdge\n");
  Serial.printf("\tNode\tNode\tFace\tFace\tType\n");
  for(i=0;i<nume;i++)
  {
    Serial.printf("%4d\t%4d\t%4d\t%4d\t%4d\t%s\n",i,edgeno[i][0],
      edgeno[i][1],edgefa[i][0],edgefa[i][1],
      (concav[i]>0.0)?"concave":"convex");
  }

  Serial.printf("\n\nGalimbr has computed %d inward unit normals to the faces:\n",
    numf-icopla);
  Serial.printf("Face #\tX  Y  Z\n");
  for(i=0;i<numf-icopla;i++)
  {
    char jx='+';
    char jy='+';
    char jz='+';
    if(normal[i][0]<0.0)jx='-';
    if(normal[i][1]<0.0)jy='-';
    if(normal[i][2]<0.0)jz='-';
    Serial.printf("%4d\t%c  %c  %c\n",i,jx,jy,jz);
  }

  //Data error checking
  Serial.printf("\n\nError summary:\n\n");
  k=0;
  for(i=0;i<nume;i++)
  {
    if(edgefa[i][0]==EOF || edgefa[i][1]==EOF)
    {
      Serial.printf("Edge %d does not have 2 corresponding faces.\n",i);
      k++;
    }
    for(j=i+1;j<nume;j++)
    {
      if(edgeno[j][0]==edgeno[i][0]&&
         edgeno[j][1]==edgeno[i][1])
      {
        Serial.printf("Edge from node %d to node %d is multiply",
          edgeno[i][0],edgeno[i][1]);
        Serial.printf(" defined.\n");
        k++;
        break;
      }
    }
  }
  if(k)
  {
    Serial.printf("\n%d Errors detected - Galimbr aborted!!",k);
    exit(1);
  }
}

//Private member function to capture certain errors
void galimber::error(char *errclass) //Print error class
{
  Serial.printf("Error class: %s\n", errclass);
  exit(1);
}

//Private member function to compute direction cosine matrix to affect rotation
void galimber::rotate(double psi, double theta, double phi, double dircos[3][3])
{
  register double spsi=sin(psi);
  register double cpsi=cos(psi);
  register double sthe=sin(theta);
  register double cthe=cos(theta);
  register double sphi=sin(phi);
  register double cphi=cos(phi);
  dircos[0][0]=cthe*cpsi;
  dircos[0][1]= -cthe*spsi;
  dircos[0][2]=sthe;
  dircos[1][0]= -sthe*sphi*cpsi-cphi*spsi;
  dircos[1][1]=sthe*sphi*spsi-cphi*cpsi;
  dircos[1][2]=cthe*sphi;
  dircos[2][0]=sthe*cphi*cpsi-sphi*spsi;
  dircos[2][1]= -sthe*cphi*spsi-sphi*cpsi;
  dircos[2][2]= -cthe*cphi;
}

//Private member function to rotate a single point
void galimber::xformr(double in[3], double dircos[3][3], double out[3])
{
  out[0]=dircos[0][0]*in[0]+dircos[0][1]*in[1]+dircos[0][2]*in[2];
  out[1]=dircos[1][0]*in[0]+dircos[1][1]*in[1]+dircos[1][2]*in[2];
  out[2]=dircos[2][0]*in[0]+dircos[2][1]*in[1]+dircos[2][2]*in[2];
}

//Do HLE projection - this is the member function that does all the work.
//Inputs are the 3 rotation angles.
//Output is a queue of line segments to be drawn.
void galimber::project(double ipsi, double ithe, double iphi,
                       deque<line> &linequeue)
{
  int i,j,k,m; //Persistent loop indices

  //Compute direction cosine matrix for this projection.
  double psi = ipsi/180.0*M_PI + 0.0017; //To radians with slight perturbation
  double theta = -ithe/180.0*M_PI + 0.0017;
  double phi = -iphi/180.0*M_PI + 0.0017;
  double dircos[3][3];
  rotate(psi,theta,phi,dircos);

  //Initialize main loop variables.
  for(i=0; i<nume; i++) intensity[i] = 0xfff;
  for(i=0; i<nume; i++) for(j=0; j<20; j++) vnatur[i][j] = 0;
  for(i=0;i<200;i++) isecnt[i] = 0;

  //Transform all nodes to verticies by rotation.
  for(i=0; i<nump; i++)
  {
    xformr(p[i], dircos, v[i]);
    //Update maxdim on the fly.
    if(fabs(v[i][1]) > maxdim) maxdim = fabs(v[i][1]);
    if(fabs(v[i][2]) > maxdim) maxdim = fabs(v[i][2]);
  }

  //For each face, compute onview factor.
  icopla=0;
  for(k=0;k<numf;k++)
  {
    if(face[k][19])
    {
      icopla++;
    }
    else
    {
      if((v[face[k][1]][1]-v[face[k][0]][1])*
         (v[face[k][2]][2]-v[face[k][0]][2])-
         (v[face[k][1]][2]-v[face[k][0]][2])*
         (v[face[k][2]][1]-v[face[k][0]][1])>=0.)
      {
        onview[k-icopla]=0;
      }
      else
      {
        onview[k-icopla]=1;
      }
    }
  }

  //Eliminate edges hidden by own volume.
  for(i=0;i<nume;i++)
  {
    if(onview[edgefa[i][0]]==0 && onview[edgefa[i][1]]==0)
    {
      intensity[i]=0xffe;
    }
    else
    {
      if(concav[i]>0.0 && (onview[edgefa[i][0]]==0 ||
         onview[edgefa[i][1]]==0))
      {
        intensity[i]=0xffd;
      }
    }
  }

  //Outer loop to calculate all intersections between remaining visible edges
  for(i=0;i<nume-1;i++)
  {
    register double *index;
    register double *jndex;
    if(intensity[i]==0xffe)continue;

    index=v[edgeno[i][0]];
    jndex=v[edgeno[i][1]];

    //Inner edge intersection loop
    for(j=i+1;j<nume;j++)
    {
      register double *kndex;
      register double *lndex;
      register double fact1;
      register double fact2;
      register double xintr;
      register double yintr;
      register double fract1;
      register double fract2;
      register double p1;
      register double p2;
      
      if(intensity[j]==0xffe)continue;

      kndex=v[edgeno[j][0]];
      lndex=v[edgeno[j][1]];

      //If edges share a node, skip.
      if(index==kndex||index==lndex||jndex==kndex||
         jndex==lndex)continue;

      fact2=lndex[1]-kndex[1];
      if(dabs(fact2)>0.0001)
      {
        fact2=(kndex[2]-lndex[2])/fact2;
      }
      else
      {
        fact2=1.0E+99;
      }
      fact1=jndex[1]-index[1];
      if(dabs(fact1)>0.0001)
      {
        fact1=(index[2]-jndex[2])/fact1;
      }
      else
      {
        fact1= -1.0E+99;
      }

      //If parallel, skip.
      if(dabs(fact1-fact2)<0.0001)continue;

      //X intersection by cramers rule.
      if(fact2 != 1.0E+99)
      {
        if(fact1 != -1.0E+99)
        {
          xintr=(kndex[2]+fact2*
           kndex[1]-index[2]-
           fact1*index[1])/
           (fact2-fact1);
        }
        else
        {
          xintr=index[1];
        }
      }
      else
      {
        xintr=kndex[1];
      }

      //If x intersection not on both segments, skip.
      if(xintr<dmin(index[1],jndex[1]) ||
         xintr<dmin(kndex[1],lndex[1]) ||
         xintr>dmax(index[1],jndex[1]) ||
         xintr>dmax(kndex[1],lndex[1]))continue;

      //Y intersection by substitution.
      if(fact1==-1.0E+99)
      {
        yintr= -fact2*(xintr-kndex[1])+
               kndex[2];
      }
      else
      {
        yintr= -fact1*(xintr-index[1])+
               index[2];
      }

      //If y intersection not on both segments, skip.
      if(yintr<dmin(index[2],jndex[2]) ||
         yintr<dmin(kndex[2],lndex[2]) ||
         yintr>dmax(index[2],jndex[2]) ||
         yintr>dmax(kndex[2],lndex[2]))continue;

      //Fractional distance along first segment.
      fract1=jndex[1]-index[1];
      if(dabs(fract1)>0.01)
      {
        fract1=(xintr-index[1])/fract1;
      }
      else
      {
        fract1=(yintr-index[2])/
               (jndex[2]-index[2]);
      }

      //Transformed LOS coordinate of first segment intersection.
      p1=(jndex[0]-index[0])*fract1+index[0];

      //Fraction along second segment.
      fract2=lndex[1]-kndex[1];
      if(dabs(fract2)>0.01)
      {
        fract2=(xintr-kndex[1])/fract2;
      }
      else
      {
        fract2=(yintr-kndex[2])/
               (lndex[2]-kndex[2]);
      }

      //Transformed LOS coordinate of second segment intersection.
      p2=(lndex[0]-kndex[0])*fract2+kndex[0];

      //Which segment is losing?
      if(p1>p2)
      {
        //Store for second segment losing.
        if(fract2<=0.001) continue;
        if(isecnt[j]>20)error("2");
        pintse[j][isecnt[j]]=fract2;
        eintse[j][isecnt[j]++]=(short)i;
      }
      else
      {
        //Store for first segment losing.
        if(fract1<=0.001) continue;
        if(isecnt[j]>20)error("2");
        pintse[i][isecnt[i]]=fract1;
        eintse[i][isecnt[i]++]=(short)j;
      }
    }
  }

  //Compute nature of point near initial vertex for each remaining edge.
  for(i=0;i<nume;i++)
  {
    register double *indexp;
    register double *jndexp;
    register double *kndexp;
    register double *lndexp;
    register double xintr;
    register double yintr;
    register double alt;
    register int jndex;
    register int kndex;
    register int ivnat;

    if(intensity[i]==0xffe || intensity[i]==0xffd)continue;
    indexp=v[edgeno[i][0]];
    lndexp=v[edgeno[i][1]];
    xintr=indexp[1]+(lndexp[1]-indexp[1])*0.001;
    yintr=indexp[2]+(lndexp[2]-indexp[2])*0.001;
    alt=  indexp[0]+(lndexp[0]-indexp[0])*0.001;
    for(j=0;j<numf;j++)face[j][20]=0;

    //Determine which polygons onview point is inside of.
    for(j=0;j<nume;j++)
    {
      register double a;

      //If corresponding polygons not onview, skip.
      if(onview[edgefa[j][0]]==0 &&
         onview[edgefa[j][1]]==0) continue;

      jndexp=v[edgeno[j][0]];
      kndexp=v[edgeno[j][1]];

      //If y coordinate of point not in edge, skip.
      if(yintr<dmin(jndexp[2],kndexp[2]) ||
         yintr>dmax(jndexp[2],kndexp[2]))continue;
      a=kndexp[2]-jndexp[2];
      if(dabs(a)>0.001)
      {
        //If point not to right of edge, skip.
        if((yintr-jndexp[2])/a*(kndexp[1]-
           jndexp[1])+jndexp[1]>xintr)
           continue;
      }

      //If first onview face on edge of initial point, skip it.
      jndex=edgefa[j][0];
      kndex=edgefa[j][1];
      if(onview[jndex] != 0)
      {
        if(edgefa[i][0]!=jndex &&
          edgefa[i][1]!=jndex)face[jndex][20]++;
      }

      //If second onview face on edge of initial point, skip it.
      if(onview[kndex] != 0)
      {
        if(edgefa[i][0]!=kndex &&
          edgefa[i][1]!=kndex)face[kndex][20]++;
      }
    }

    //If face[-][20] is odd, point lies in that polygon.
    ivnat=0;

    //Loop to compile which faces hide point in vnatur*/
    icopla=0;
    for(j=0;j<numf;j++)
    {
      if(face[j][19])
      {
        icopla++;
      }
      else
      {
        if(face[j-icopla][20]&1)
        {
          register double werb,xerb;
          register double yerb,zerb;
          register double s;
          register double t;
          register double denom;
          indexp=v[face[j][0]];
          jndexp=v[face[j][1]];
          kndexp=v[face[j][2]];

          //Compute LOS coord of LOS-plane intersection.
          werb=jndexp[2]-indexp[2];
          xerb=kndexp[1]-indexp[1];
          yerb=jndexp[1]-indexp[1];
          zerb=kndexp[2]-indexp[2];
          denom=yerb*zerb-xerb*werb;
          if(denom==0.0)continue;
          s=((xintr-indexp[1])*zerb-
             (yintr-indexp[2])*xerb);
          t=((yintr-indexp[2])*yerb-
             (xintr-indexp[1])*werb);
          if(s!=0.0)s=s/denom;
          if(t!=0.0)t=t/denom;

          //If LOS coordinate of intersection behind point, skip.
          if(alt<=(indexp[0]+
           s*(jndexp[0]-indexp[0])+
           t*(kndexp[0]-indexp[0])))
          {
            if(ivnat>20)error("1");
            vnatur[i][ivnat++]=
             (short)(j-icopla+1);
          }
        }
      }
    }
  }

  //Initial nature of all remaining edges known, print debug if debug flag set.
  if(debug)
  {
    Serial.printf("\n\nDEBUG PRINT:\n\n");
    Serial.printf("When PSI   is %f\n",psi);
    Serial.printf(" and THETA is %f\n",theta);
    Serial.printf(" and PHI   is %f\n",phi);
    Serial.printf("\nThe %d vertices are:\n\n",nump);
    for(i=0;i<nump;i++)
    {
      Serial.printf("%10.2f%10.2f%10.2f\n",
       v[i][0],v[i][1],v[i][2]);
    }
    Serial.printf("\n\nThe %d edge intensities are:\n\n",
     nume);
    for(i=0;i<nume;i++)
    {
      Serial.printf("%5o",intensity[i]);
      if((i+1)%10==0)Serial.printf("\n");
    }
    Serial.printf("\n");
    Serial.printf("\n\nThe %d face onview factors are:\n\n",numf-icopla);
    for(i=0;i<numf-icopla;i++)
    {
      Serial.printf("%5d",onview[i]);
      if((i+1)%10==0)Serial.printf("\n");
    }
    Serial.printf("\n");
    Serial.printf("\n\nThe %d edge intersections counts are:\n\n",nume);
    for(i=0;i<nume;i++)
    {
      Serial.printf("%5d",isecnt[i]);
      if((i+1)%10==0)Serial.printf("\n");
    }
    Serial.printf("\n");
    Serial.printf("\n\nThe edge intersection lists are:\n\n");
    for(i=0;i<nume;i++)
    {
      for(j=0;j<20;j++)
      {
        Serial.printf("%5.3f",pintse[i][j]);
      }
      Serial.printf("\n");
    }
    Serial.printf("\n\nThe intersecting segment lists are:\n\n");
    for(i=0;i<nume;i++)
    {
      for(j=0;j<20;j++)
      {
        Serial.printf("%4d",eintse[i][j]);
      }
      Serial.printf("\n");
    }
    Serial.printf("\n\nThe edge initial vertex nature lists are:\n\n");
    for(i=0;i<nume;i++)
    {
      for(j=0;j<20;j++)
      {
        Serial.printf("%4d",vnatur[i][j]);
      }
      Serial.printf("\n");
    }
  }//End debug print

  //Generate output vector list in linequeue.
  for(i=0;i<nume;i++)
  {
    register int l;
    register double *index;
    register double *jndex;
    register int ii;
    register int mm;
    register int nature=0;
    register double a;
    char *ivis;
    double from_vertex_x;
    double from_vertex_y;
    double from_vertex_z;

    //Draw no edges hidden by own volume.
    if(intensity[i]==0xffe || intensity[i]==0xffd)continue;

    //See if all vnatur zero.
    for(l=0;l<20;l++) nature+=vnatur[i][l];

    //If initial vertex hidden and no intersections, do not draw edge.
    if((mm=isecnt[i])==0 && nature)continue;

    //Move to initial point on edge.
    index=v[edgeno[i][0]];
    jndex=v[edgeno[i][1]];
    from_vertex_x = index[1];
    from_vertex_y = index[2];
    from_vertex_z = index[0];

    //Main loop for individual segments of edge.
    for(;;)
    {
      ivis="l";
      nature=0;
      for(l=0;l<20;l++)nature+=vnatur[i][l];
      if(nature!=0)ivis="m";
      if(isecnt[i])
      {
        //Still more intersections - determine lowest fractional remaining one.
        a=100.0;
        for(k=0;k<mm;k++)
        {
          if(pintse[i][k]<a)
          {
            a=pintse[i][k];
            ii=k;
          }
        }

        //Draw vector to this point visibility dependent on vnatur.
        double x=(jndex[1]-index[1])*a+index[1];
        double y=(jndex[2]-index[2])*a+index[2];
        double z=(jndex[0]-index[0])*a+index[0];
        if(ivis[0]=='m')
        {
          //Move to vertex.
          from_vertex_x = x;
          from_vertex_y = y;
          from_vertex_z = z;
        }
        else
        {
          //Queue line segment to linequeue.
          //linequeue.push_back(line( short(from_vertex_x/(maxdim)*64+80),
          //                          short(-from_vertex_y/(maxdim)*64+64),
          //                          short(x/(maxdim)*64+80),
          //                          short(-y/(maxdim)*64+64) ));
         //Queue line segment to linequeue, scaled to +/- 1.0.
          linequeue.push_back(line( from_vertex_x/maxdim,
                                    -from_vertex_y/maxdim,
                                    x/maxdim,
                                    -y/maxdim ));
          from_vertex_x = x;
          from_vertex_y = y;
          from_vertex_z = z;
        }

        isecnt[i]--;
        pintse[i][ii]=100.0;

        //Update vnatur as cross intersection.
        m=edgefa[eintse[i][ii]][0];
        if(onview[m])
        {
          for(l=0;l<20;l++)
          {
            if(m+1==vnatur[i][l])
            {
              vnatur[i][l]=0;
              break;
            }
          }
          for(l=(l!=20)*21;l<20;l++)
          {
            if(vnatur[i][l]==0)
            {
              vnatur[i][l]=
               (short)(m+1);
              break;
            }
          }
          if(l==20)error("708");
        }
        m=edgefa[eintse[i][ii]][1];
        if(onview[m])
        {
          for(l=0;l<20;l++)
          {
            if(m+1==vnatur[i][l])
            {
              vnatur[i][l]=0;
              break;
            }
          }
          for(l=(l!=20)*21;l<20;l++)
          {
            if(vnatur[i][l]==0)
            {
              vnatur[i][l]=
               (short)(m+1);
              break;
            }
          }
          if(l==20)error("708");
        }
      }
      else
      {
        //No more intersections - draw lst* segment only if visible.
        if(nature==0)
        {
          //Queue line segment to linequeue.
          //linequeue.push_back(line( short(from_vertex_x/(maxdim)*64+80),
          //                          short(-from_vertex_y/(maxdim)*64+64),
          //                          short(jndex[1]/(maxdim)*64+80),
          //                          short(-jndex[2]/(maxdim)*64+64) ));
         //Queue line segment to linequeue, scaled to +/- 1.0.
          linequeue.push_back(line( from_vertex_x/maxdim,
                                    -from_vertex_y/maxdim,
                                    jndex[1]/maxdim,
                                    -jndex[2]/maxdim ));
        }
        break;
      }
    }//End loop drawing segments of an edge.
  }//End loop drawing edges.
}
