/*  This program reads one of several 3-dimensional object description files and
 *  renders on the TFT display an orthogonal projection (or 2-dimensional view
 *  of the object from a distant viewpoint) of that object after it has been
 *  rotated about 3 axes affixed at the center-of-gravity of the object. The
 *  angles of rotation are incremented after each projection such that the
 *  object appears to rotate in space as the projections are updated on the
 *  display in real-time. The projection does not provide for perspective, as
 *  with a viepoint or point source of light close to the object.
 *  
 *  The projection eliminates "hidden lines", that is, edges or portions thereof
 *  that could not be seen through the solid parts of the object because they
 *  are hidden by the volume of the object.
 *  
 *  The objects are described by a number of points (nodes) and a number of
 *  planar faces that result from edges connecting those nodes. This object
 *  model cannot represent "warped" faces; every face must lie in a plane.
 *  Since the object is "solid", each face has an outside and inside. The
 *  outside of the face is described by following its edges from node to node
 *  keeping the face to the right. Thus an object (or part of an object) with no
 *  thickness must be described by 2 faces, both viewed from the "outside".
 *  Also, a face may have "holes" in it and must be described as several lists
 *  of nodes (a single face with several coplaner edge lists). There may be up
 *  to 200 points and 200 faces, and each edge list describing a face boundary
 *  (or coplaner face boundary) may contain up to 18 points. The object files
 *  are simple text files; points are numbered starting with 0 and are
 *  specified in an arbitrarily scaled floating point 3-D coordinate system.
 *  
 *  This program uses an algorithm described in a paper from Communications of
 *  the ACM, Vol. 12, No. 4, April 1969 entitled "An Algorithm for Hidden Line
 *  Elimination" authored by R. Galimberti and U. Montanari of Milan, Italy.
 *  The implementation of this algorithm is wholly contained within the source
 *  files galimber.h and galimber.cpp, which define the class "galimber".
 *  
 *  Ported to Ringo/Makerphone 7 Apr 2020 Frank Prindle (Version 1.0)
 *  Added minor fix to improve exit from program 8 Apr 2020 Frank Prindle (Version 1.0.1)
 *  
 *  Use left function button to decrease increment for angle psi
 *  Use power button to increase increment for angle psi
 *  Use home button to decrease increment for angle theta
 *  Use right function button to increase increment for angle theta
 *  Use asterisk button to decrease increment for angle phi
 *  Use pound (AKA hashtag) button to increase increment for angle phi
 *  Use A button to restart program
 *  Use B button to quit program
 */
 
#include <string>
#include <sstream>
#include "galimber.h"
#include <MAKERphone.h>
using namespace std;

galimber *gp = 0; //Pointer to a galimber hidden line eliminator
MAKERphone mp; //Makerphone library provides access to display and buttons
double ipsi, ithe, iphi; //Rotation angles psi, theta, and phi in degrees
double ipsi_incr, ithe_incr, iphi_incr; //Angle increments in degrees

//Initialization
void setup()
{
  //Allow for restart
  if(gp)
  {
    //Restart case... just delete old galimber hidden line eliminator
    delete gp;
  }
  else
  {
    //Initial startup
    Serial.begin(115200);
    mp.begin(1);    
    mp.inCall=1; //Keep mp awake for straightforward exit (FCP 4/8/2020)
  }

  //Display the object menu.
  mp.display.setTextColor(TFT_YELLOW);
  mp.display.fillScreen(TFT_BLACK);
  mp.display.setCursor(0,0);
  mp.display.print("  Select object number\n      1. SIMPLE\n      2. TRIANGLE\n      "
                   "3. SAMPLE\n      4. EXPLODED\n      5. CIGAR\n      6. FIGHTER\n      "
                   "7. HELO\n\n");
  mp.display.pushSprite(0,0);

  //Wait for user to select an object using buttons 1-7.
  int object = 0;
  while(!object)
  {
    mp.buttons.update();
    for(int i=0; i<7; i++)
    {
      if(mp.buttons.released(i))
      {
        object = i+1;
        break;
      }
    }
    if(mp.buttons.released(BTN_B)){mp.update();mp.loader();}//B reboots firmware if available
  }

  //Set equivalent object file name (files must exist in directory "/galimber").
  char *filename;
  if(object == 1) filename="/galimber/simple.obj";
  if(object == 2) filename="/galimber/triangle.obj";
  if(object == 3) filename="/galimber/sample.obj";
  if(object == 4) filename="/galimber/exploded.obj";
  if(object == 5) filename="/galimber/cigar.obj";
  if(object == 6) filename="/galimber/fighter.obj";
  if(object == 7) filename="/galimber/helo.obj";
  
  //Read in the object definition file.
  File objectdef = SD.open(filename,"r");
  char file[objectdef.size()+1];
  objectdef.read((unsigned char*)file,objectdef.size());
  file[objectdef.size()] = '\0'; //Make it a proper null-terminated string

  //Parse the object file into points and edges.
  string s(file);
  istringstream is(s);
  int nump;
  is >> nump; //First line is number of points
  static double p[200][3]; //Allocate statically - too big for stack otherwise
  for(int i=0;i<nump;i++) is >> p[i][0] >> p[i][1] >> p[i][2]; //x,y,z per line
  int numf;
  is >> numf; //Next line is number of faces
  static double face[200][20]; //Allocate statically - too big for stack otherwise
  for(int i=0;i<numf;i++)
  {
    for(int j=0;j<19;j++)
    {
      is >> face[i][j]; //List of points with face to the right
      if(face[i][j] == EOF) break; //List ends with a -1
    }
    //There may be a 'c' at the end of each line (coplaner list for prior face)
    char remainder[10];
    is.getline(remainder,10);
    if(remainder[0] == '\0' || remainder[1] == '\0' || remainder[2] == '\0')
    {
      face[i][19] = 0; //not coplaner
    }
    else
    {
      face[i][19] = 1; //coplaner (flag list as such in its 20th element).
    }
  }
  
  //Allocate a galimber hidden line eliminator
  //for the object described by the file (pass in points and faces).
  gp = new galimber(nump,p,numf,face);
  
  ipsi = ithe = iphi = 0.0; //Initialize rotation angles
  ipsi_incr = 0.5/0.62; ithe_incr = 0.7/0.82; iphi_incr = 0.3/1.02; //Initialize angle increments
  if(object>5) //More complex objects render slower, need larger increments
  {
    ipsi_incr *= (object-4);
    ithe_incr *= (object-4);
    iphi_incr *= (object-4);
  }
}

//Generate and display a single projection.
void loop()
{
  galimber &g = *gp; //Reference to the galimber hidden line eliminator
  static deque<line> linequeue; //Queue to receive line segments to be drawn

  //HLE project defined object rotated about 3 axes by ipsi, ithe, and iphi.
  //Line segments are returned in linequeue.
  g.project(ipsi, ithe, iphi, linequeue);

  //Draw the generated line segments onto 160 x 128 display canvas.
  mp.display.fillScreen(TFT_BLACK);
  while(linequeue.size())
  {
    line &l = linequeue.front();
    mp.display.drawLine(int(l.xstart*64+80), int(l.ystart*64+64),
                        int(l.xend*64+80), int(l.yend*64+64), TFT_GREEN);
    linequeue.pop_front();
  }
  mp.display.pushSprite(0,0); //Push display to TFT screen

  //Now increment the angles.
  ipsi += ipsi_incr;
  ithe += ithe_incr;
  iphi += iphi_incr;

  //Allow top row of buttons and asterisk and pound/hashtag to increase/decrease
  //the increments by a small amount per button press.
  #define BTN_POWER 14
  mp.buttons.update();
  if(mp.buttons.released(BTN_FUN_LEFT)) ipsi_incr -= 0.3;
  if(mp.buttons.released(BTN_POWER)) ipsi_incr += 0.3;
  if(mp.buttons.released(BTN_HOME)) ithe_incr -= 0.3;
  if(mp.buttons.released(BTN_FUN_RIGHT)) ithe_incr += 0.3;
  if(mp.buttons.released(BTN_ASTERISK)) iphi_incr -= 0.3;
  if(mp.buttons.released(BTN_HASHTAG)) iphi_incr += 0.3;
  
  //Quit on B button; restart on the A button.
  if(mp.buttons.released(BTN_B)) {mp.update();mp.loader();} //B reboots firmware if available
  if(mp.buttons.released(BTN_A)) setup(); // A restarts
}
